@@ -1,4 +1,4 @@
-"""a collection of helper-functions to generate map-plots"""
+"""A collection of helper-functions to generate map-plots."""
 
 from functools import partial, lru_cache, wraps
 from collections import defaultdict
@@ -10,9 +10,11 @@ import numpy as np
 
 try:
     import geopandas as gpd
-except:
-    print("geopandas could not be imported... 'add_overlay' not working!")
-    pass
+
+    _gpd_OK = True
+except ImportError:
+    _gpd_OK = False
+
 from scipy.spatial import cKDTree
 from pyproj import CRS, Transformer
 
@@ -27,7 +29,12 @@ from cartopy import crs as ccrs
 from cartopy import feature as cfeature
 from cartopy.io import shapereader
 
-from .helpers import pairwise, cmap_alpha, BlitManager, draggable_axes
+from .helpers import (
+    pairwise,
+    cmap_alpha,
+    BlitManager,
+    draggable_axes,
+)
 from ._shapes import shapes
 
 from ._containers import (
@@ -40,7 +47,7 @@ from ._containers import (
 )
 
 from ._cb_container import cb_container
-
+from .scalebar import ScaleBar
 
 try:
     import mapclassify
@@ -50,7 +57,7 @@ except ImportError:
 
 class Maps(object):
     """
-    The base-class for generating plots with EOmaps
+    The base-class for generating plots with EOmaps.
 
     Note: if you want to plot a grid of maps, checkout `MapsGrid`!
 
@@ -68,6 +75,9 @@ class Maps(object):
         >>> m1 = Maps()
         >>> m2 = Maps(parent=m1)
 
+        Note: Instead of specifying explicit axes, you might want to have a
+        look at `eomaps.MapsGrid` objects!
+
     orientation : str, optional
         Indicator if the colorbar should be plotted right of the map
         ("horizontal") or below the map ("vertical"). The default is "vertical"
@@ -213,8 +223,15 @@ class Maps(object):
         self._ax = gs_ax
         self._init_ax = gs_ax
 
-        # max. pick radius for pick-events
-        self._pick_distance = None
+    def _check_gpd(self):
+        # raise an error if geopandas is not found
+        # (execute this in any function that requires geopandas!)
+        if not _gpd_OK:
+            raise ImportError(
+                "EOmaps: You need to install geopandas first!\n"
+                + "... with conda, simply use:  "
+                + "'conda install -c conda-forge geopandas'"
+            )
 
     @property
     @lru_cache()
@@ -267,23 +284,29 @@ class Maps(object):
                     plt.ioff()
                 else:
                     plt.ion()
+            else:
+                # make sure the parent Maps-object is initialized as well!
+                self.parent._set_axes()
             plt.show()
 
-    # def _reset_axes(self):
-    #     print("resetting")
-    #     for m in [self.parent, *self.parent._children]:
-    #         m._f = None
-    #         m._ax = m._init_ax
-    #         m._ax_cb = None
-    #         m._ax_cb_plot = None
-    #         m._gridspec = None
-    #         m._cb_gridspec = None
+    def _reset_axes(self):
+        # reset all objects connected to the figure (in case it is closed)
+        for m in [self.parent, *self.parent._children]:
+            m._f = None
+            m._ax = m._init_ax
+            m._ax_cb = None
+            m._ax_cb_plot = None
+            m._gridspec = None
+            m._cb_gridspec = None
 
-    #     # # reset the Blit-Manager
-    #     self.parent._BM = None
+            # reset all callbacks attached to the figure
+            m.cb._reset_cids()
 
-    #     # reset the draggable-axes class on next call
-    #     self.parent._axpicker = None
+        # reset the Blit-Manager
+        self.parent._BM = None
+
+        # reset the draggable-axes class on next call
+        self.parent._axpicker = None
 
     @property
     def BM(self):
@@ -435,7 +458,7 @@ class Maps(object):
             assert (
                 "parent" not in kwargs
             ), "EOmaps: parent is set automatically if you use 'connect=True'"
-            kwargs["parent"] = self
+            kwargs["parent"] = self.parent
 
         # create a new class
         copy_cls = Maps(**kwargs)
@@ -648,8 +671,7 @@ class Maps(object):
             The classification scheme to use.
             (the list is accessible via `m.classify_specs.SCHEMES`)
 
-            E.g. one of:
-            [ "scheme" (\**kwargs)  ]
+            E.g. one of (possible kwargs in brackets):
 
                 - BoxPlot (hinge)
                 - EqualInterval (k)
@@ -667,55 +689,12 @@ class Maps(object):
                 - StdMean (multiples)
                 - UserDefined (bins)
 
-        \**kwargs :
+        kwargs :
             kwargs passed to the call to the respective mapclassify classifier
+            (dependent on the selected scheme... see above)
         """
         self.classify_specs._set_scheme_and_args(scheme, **kwargs)
 
-    def _pick_pixel(self, artist, event):
-        if self._pick_distance is None:
-            maxdist = np.inf
-        else:
-            maxdist = self._pick_distance
-
-        if (event.inaxes != self.figure.ax) or not hasattr(self, "tree"):
-            return True, dict(ind=None, dblclick=event.dblclick, button=event.button)
-
-        # use a cKDTree based picking to speed up picks for large collections
-        dist, index = self.tree.query((event.xdata, event.ydata))
-        # set max. distance in pixel-coordinates for picking
-
-        # do this to make sure that we calculate the distance in the right axes!
-        # (necessary for connected pick-callbacks)
-        # for the axes of the original click, this is equal to (event.x, event.y)
-        p1 = self.figure.ax.transData.transform((event.xdata, event.ydata))
-
-        p2 = self.figure.ax.transData.transform(
-            (self._props["x0"][index], self._props["y0"][index])
-        )
-        pdist = np.sqrt(np.sum((p1 - p2) ** 2))
-        # print(dist, index, pdist, maxdist, pdist < maxdist, index)
-
-        if pdist < maxdist:
-            self._pick = True, dict(
-                ind=index, dblclick=event.dblclick, button=event.button, dist=dist
-            )
-            return True, dict(
-                ind=index, dblclick=event.dblclick, button=event.button, dist=dist
-            )
-        else:
-            return True, dict(
-                ind=None, dblclick=event.dblclick, button=event.button, dist=dist
-            )
-
-        return False, None
-
-    def _attach_picker(self, picker=None):
-        if picker is None:
-            self.figure.coll.set_picker(self._pick_pixel)
-        else:
-            self.figure.ax.set_picker(picker)
-
     @property
     @lru_cache()
     def _bounds(self):
@@ -1127,7 +1106,7 @@ class Maps(object):
             n_cmap,
             cax=ax_cb,
             label=label,
-            extend="both",
+            extend="neither",
             spacing="proportional",
             orientation=cb_orientation,
         )
@@ -1286,7 +1265,7 @@ class Maps(object):
 
         return cb
 
-    def add_gdf(self, gdf, **kwargs):
+    def add_gdf(self, gdf, picker_name=None, pick_method="contains", **kwargs):
         """
         Overplot a `geopandas.GeoDataFrame` over the generated plot.
         (`plot_map()` must be called!)
@@ -1295,17 +1274,101 @@ class Maps(object):
         ----------
         gdf : geopandas.GeoDataFrame
             A GeoDataFrame that should be added to the plot.
-        **kwargs :
-            all kwargs are passed to `gdf.plot(**kwargs)`
+        picker_name : str or None
+            A unique name that is used to identify the pick-method.
+
+            If a `picker_name` is provided, a new pick-container will be
+            created that can be used to pick geometries of the GeoDataFrame.
+
+            The container can then be accessed via:
+                >>> m.cb.pick__<picker_name>
+                or
+                >>> m.cb.pick[picker_name]
+            and it can be used in the same way as `m.cb.pick...`
+
+        pick_method : str or callable
+            if str :
+                The operation that is executed on the GeoDataFrame to identify
+                the picked geometry
+            if callable :
+                A callable that is used to identify the picked geometry.
+                The call-signature is:
+
+                >>> def picker(artist, mouseevent):
+                >>>     # if the pick is NOT successful:
+                >>>     return False, dict()
+                >>>     ...
+                >>>     # if the pick is successful:
+                >>>     return True, dict(ID, pos, val, ind)
+            The default is "contains"
+        kwargs :
+            all remaining kwargs are passed to `gdf.plot(**kwargs)`
         """
+        self._check_gpd()
+
         self._set_axes()
-
         ax = self.figure.ax
-
         defaultargs = dict(facecolor="none", edgecolor="k", lw=1.5)
         defaultargs.update(kwargs)
 
-        gdf.to_crs(self.crs_plot).plot(ax=ax, aspect=ax.get_aspect(), **defaultargs)
+        # transform data to the plot crs
+        usegdf = gdf.to_crs(self.crs_plot)
+
+        # plot gdf and identify newly added collections
+        # (geopandas always uses collections)
+        colls = [id(i) for i in self.figure.ax.collections]
+        usegdf.plot(ax=ax, aspect=ax.get_aspect(), **defaultargs)
+        newcolls = [i for i in self.figure.ax.collections if id(i) not in colls]
+        if len(newcolls) > 1:
+            prefixes = [
+                f"_{i.__class__.__name__.replace('Collection', '')}" for i in newcolls
+            ]
+            warnings.warn(
+                "EOmaps: Multiple geometry types encountered in `m.add_gdf`. "
+                + "The pick containers are re-named to"
+                + f"{[picker_name + prefix for prefix in prefixes]}"
+            )
+        else:
+            prefixes = [""]
+
+        if picker_name is not None:
+            if pick_method is not None:
+                if isinstance(pick_method, str):
+
+                    def picker(artist, mouseevent):
+                        try:
+                            query = getattr(usegdf, pick_method)(
+                                gpd.points_from_xy(
+                                    [mouseevent.xdata], [mouseevent.ydata]
+                                )[0]
+                            )
+                            if query.any():
+                                inds = usegdf.index[query]
+                                return True, dict(
+                                    ind=inds[0],
+                                    dblclick=mouseevent.dblclick,
+                                    button=mouseevent.button,
+                                )
+                            else:
+                                return False, dict()
+                        except:
+                            return False, dict()
+
+                elif callable(pick_method):
+                    picker = pick_method
+                else:
+                    print(
+                        "EOmaps: I don't know what to do with the provided pick_method"
+                    )
+
+                # explode the GeoDataFrame to avoid picking multi-part geometries
+                usegdf = usegdf.explode(index_parts=False)
+                for art, prefix in zip(newcolls, prefixes):
+                    # make the newly added collection pickable
+                    self.cb.add_picker(picker_name + prefix, art, picker=picker)
+
+                    # attach the re-projected GeoDataFrame to the pick-container
+                    self.cb.pick[picker_name + prefix].data = usegdf
 
     def add_coastlines(self, layer=None, coast=True, ocean=True):
         """
@@ -1376,8 +1439,8 @@ class Maps(object):
         Check `cartopy.shapereader.natural_earth` for details on how to specify
         layer properties.
 
-        to change the appearance and position of the map (or to add it to the
-        plot at a later stage) use
+        To get more control of the position and appearance of the legend
+        (or to add it to the plot at a later stage) use `legend=False` and explicitly call
 
             >>> m.add_overlay_legend()
 
@@ -1385,7 +1448,7 @@ class Maps(object):
         ----------
         dataspec : dict
             the data-specification used to load the data via
-            cartopy.shapereader.natural_earth(\**dataspec)
+            cartopy.shapereader.natural_earth(... dataspecs ...)
 
             >>> dataspec=(resolution='10m', category='cultural', name='urban_areas')
             >>> dataspec=(resolution='10m', category='cultural', name='admin_0_countries')
@@ -1407,6 +1470,8 @@ class Maps(object):
             a geopandas.GeoDataFrame that will be used as a mask for overlay
             (does not work with line-geometries!)
         """
+        self._check_gpd()
+
         self._set_axes()
         ax = self.figure.ax
 
@@ -1554,7 +1619,12 @@ class Maps(object):
 
             The default is None in which case "pixel" is used if a dataset is
             present and otherwise a shape with 1/10 of the axis-size is plotted
-
+        radius_crs : str or a crs-specification
+            The crs specification in which the radius is provided.
+            Either "in", "out", or a crs specification (e.g. an epsg-code,
+            a PROJ or wkt string ...)
+            The default is "in" (e.g. the crs specified via `m.data_specs.crs`).
+            (only relevant if radius is NOT specified as "pixel")
         shape : str, optional
             Indicator which shape to draw. Currently supported shapes are:
                 - geod_circles
@@ -1688,6 +1758,39 @@ class Maps(object):
         )
         self.BM.update(clear=False)
 
+    @wraps(ScaleBar.__init__)
+    def add_scalebar(
+        self,
+        lon=None,
+        lat=None,
+        azim=91.0,
+        scale=10000,
+        scale_props=None,
+        patch_props=None,
+        label_props=None,
+    ):
+
+        self._set_axes()
+
+        if lon is None and lat is None:
+            extent = self.figure.ax.get_extent()
+            lon, lat = self._transf_plot_to_lonlat.transform(
+                np.mean(extent[:2]),
+                np.mean(extent[2:]),
+            )
+
+        s = ScaleBar(
+            m=self,
+            scale=scale,
+            scale_props=scale_props,
+            patch_props=patch_props,
+            label_props=label_props,
+        )
+        s._add_scalebar(lon, lat, azim)
+        s._make_pickable()
+
+        return s
+
     @wraps(plt.savefig)
     def savefig(self, *args, **kwargs):
 
@@ -1728,7 +1831,7 @@ class Maps(object):
             Indicator if coastlines and a light-blue ocean shading should be added.
             The default is True
         pick_distance : int
-            The maximum distance (in pixels) to trigger callbacks.
+            The maximum distance (in pixels) to trigger callbacks on the added collection.
             (The distance is evaluated between the clicked pixel and the center of the
             closest data-point)
             The default is 10.
@@ -1736,17 +1839,21 @@ class Maps(object):
             A layer-index in case the collection is intended to be updated
             dynamically.
             The default is None.
-        \**kwargs
+        dynamic : bool
+            If True, the collection will be dynamically updated
+
+        kwargs
             kwargs passed to the initialization of the matpltolib collection
             (dependent on the plot-shape) [linewidth, edgecolor, facecolor, ...]
         """
+        if self.data is None and colorbar is True:
+            print("EOmaps: ...cannot add a colormap if no data is provided!")
+            colorbar = False
         self.cbQ = colorbar
 
         self._set_axes()
         ax = self.figure.ax
 
-        self._pick_distance = pick_distance
-
         if "dynamic_layer_idx" in kwargs:
             layer = kwargs.pop("dynamic_layer_idx")
             warnings.warn("EOmaps: 'dynamic_layer_idx' is depreciated... use 'layer'!")
@@ -1830,6 +1937,15 @@ class Maps(object):
 
             self.figure.coll = coll
 
+            # attach the pick-callback that executes the callbacks
+            # self.attach_picker("default", self.figure.coll, None)
+
+            # self.cb.add_picker("default", coll, None)
+            self.cb.pick._set_artist(coll)
+            self.cb.pick._init_cbs()
+            self.cb.pick._pick_distance = pick_distance
+            self.cb._methods.append("pick")
+
             if colorbar:
                 if (self.figure.ax_cb is not None) and (
                     self.figure.ax_cb_plot is not None
@@ -1844,24 +1960,16 @@ class Maps(object):
                     )
                     self.figure.cb = cb
                 else:
-                    if self.parent is self:
-                        warnings.warn(
-                            "EOmaps: Adding a colorbars is not supported if "
-                            + "you provide an explicit axes via gs_ax"
-                        )
+                    warnings.warn(
+                        "EOmaps: Colorbar could not be created... did "
+                        + "you call 'm.plot_map()' before adding other features?"
+                    )
 
             if dynamic is True:
                 self.BM.add_artist(coll, layer)
             else:
                 self.BM.add_bg_artist(coll, layer)
 
-            # ------------- add a picker that will be used by the callbacks
-            self._attach_picker()
-
-            if self.parent is self:
-                # attach the pick-callback that execute the callbacks
-                self.cb.pick._init_cbs()
-
             # set the image extent
             # get the extent of the added collection
             b = self.figure.coll.get_datalim(ax.transData)
@@ -2010,36 +2118,180 @@ class MapsGrid:
         the number of rows. The default is 2.
     c : int, optional
         the number of columns. The default is 2.
-    \**kwargs
+    m_inits, ax_inits : dict, optional
+        A dictionary that is used to initialize the Maps-objects (`m_inits`) and
+        ordinary matplotlib-axes (`ax_inits`) from the underlying GridSpec object
+
+        The keys of the dictionaries are used as names for the Maps-objects,
+        (accessible via `mgrid.m_<name>` or `mgrid.ax_<name>`) and the values
+        are used to identify the axes position from the GridSpec object.
+
+        For example, to initialize a 2 by 2 grid with a large map on top,
+        a small maps on the bottom left and a normal plot on the bottom right, use:
+
+            >>> m_inits = dict(top = (0, slice(0, 2)),
+            >>>                bottom_left=(1, 0))
+            >>> ax_inits = dict(bottom_right=(1, 1))
+
+            >>> mg = MapsGrid(2, 2, m_inits=m_inits, ax_inits=ax_inits)
+            >>> mg.m_top.plot_map()
+            >>> mg.m_bottom_left.plot_map()
+            >>> mg.ax_bottom_right.plot([1,2,3])
+
+        Note: If either `m_inits` or `ax_inits` is provided, ONLY the specified
+        objects are initialized!
+
+        The default is None in which case a unique Maps-object will be created
+        for each grid-cell (accessible via `mgrid.m_<row>_<col>`)
+
+    kwargs
         additional keyword-arguments passed to `matplotlib.gridspec.GridSpec()`
+
+    Attributes
+    ----------
+    f : matplotlib.figure
+        The matplotlib figure object
+    m_<identifier> : eomaps.Maps objects
+        The individual Maps-objects can be accessed via `mgrid.m_<identifier>`
+        The identifiers are hereby `<row>_<col>` or the keys of the `m_inits`
+        dictionary (if provided)
+    ax_<identifier> : matplotlib.axes
+        The individual (ordinary) matplotlib axes can be accessed via
+        `mgrid.ax_<identifier>`. The identifiers are hereby the keys of the
+        `ax_inits` dictionary (if provided).
+        Note: if `ax_inits` is not specified, NO ordinary axes will be created!
+
+    Methods
+    -------
+    join_limits :
+        join the axis-limits of maps that share the same projection
+    share_click_events :
+        share click-callback events between the Maps-objects
+    share_pick_events :
+        share pick-callback events between the Maps-objects
+    create_axes :
+        create a new (ordinary) matplotlib axes
+    add_<...> :
+        call the underlying `add_<...>` method on all Maps-objects of the grid
+    set_<...> :
+        set the corresponding property on all Maps-objects of the grid
+
     Returns
     -------
     eomaps.MapsGrid
-        Accessor to the Maps objects "ax_{row}_{column}".
+        Accessor to the Maps objects "m_{row}_{column}".
 
+    Notes
+    -----
+
+    - To perform actions on all Maps-objects of the grid, simply iterate over
+      the MapsGrid object!
     """
 
-    def __init__(self, r=2, c=2, **kwargs):
-        self._axes = []
+    def __init__(self, r=2, c=2, m_inits=None, ax_inits=None, **kwargs):
+        self._Maps = []
 
-        gs = GridSpec(nrows=r, ncols=c, **kwargs)
+        self._gs = GridSpec(nrows=r, ncols=c, **kwargs)
+        if m_inits is None and ax_inits is None:
+            self._custom_init = False
+            for i in range(r):
+                for j in range(c):
+                    if i == 0 and j == 0:
+                        mij = Maps(gs_ax=self._gs[0, 0])
+                        self.parent = mij
+                    else:
+                        mij = Maps(parent=self.parent, gs_ax=self._gs[i, j])
 
-        for i in range(r):
-            for j in range(c):
-                if i == 0 and j == 0:
-                    mij = Maps(gs_ax=gs[0, 0])
-                    self.parent = mij
-                else:
-                    mij = Maps(parent=self.parent, gs_ax=gs[i, j])
+                    self._Maps.append(mij)
+                    setattr(self, f"m_{i}_{j}", mij)
+        else:
+            self._custom_init = True
+            if m_inits is not None:
+                for i, [key, val] in enumerate(m_inits.items()):
+                    assert isinstance(
+                        val, tuple
+                    ), "the values of m_inits must be tuples!"
+                    assert (
+                        key.isidentifier()
+                    ), f"the provided name {key} is not a valid identifier"
 
-                self._axes.append(mij)
-                setattr(self, f"m_{i}_{j}", mij)
+                    if i == 0:
+                        mi = Maps(gs_ax=self._gs[val])
+                        self.parent = mi
+                    else:
+                        mi = Maps(parent=self.parent, gs_ax=self._gs[val])
+
+                    self._Maps.append(mi)
+                    setattr(self, f"m_{key}", mi)
+
+            if ax_inits is not None:
+                for key, val in ax_inits.items():
+                    assert isinstance(
+                        val, tuple
+                    ), "the values of ax_inits must be tuples!"
+                    assert (
+                        key.isidentifier()
+                    ), f"the provided name {key} is not a valid identifier"
+
+                    self.create_axes(val, name=key)
 
     def __iter__(self):
-        return iter(self._axes)
+        return iter(self._Maps)
 
     def __getitem__(self, key):
-        return getattr(self, f"ax_{key[0]}_{key[1]}")
+        if self._custom_init is False:
+            return getattr(self, f"m_{key[0]}_{key[1]}")
+        else:
+            return getattr(self, key)
+
+    def create_axes(self, ax_init, name=None):
+        """
+        Create (and return) an ordinary matplotlib axes.
+
+        Note: If you intend to use both ordinary axes and Maps-objects, it is
+        recommended to use an explicit "ax_inits" dict in the initialization of
+        the MapsGrid object to avoid the creation of overlapping axes!
+
+        Parameters
+        ----------
+        ax_init : set
+            The GridSpec speciffications for the axis.
+            use `ax_inits = (<row>, <col>)` to get an axis in a given grid-cell
+            use `slice(<start>, <stop>)` for `<row>` or `<col>` to get an axis
+            that spans over multiple rows/columns.
+
+        Returns
+        -------
+        ax : matplotlib.axist
+            The matplotlib axis instance
+
+        Examples
+        --------
+
+        >>> ax_inits = dict(top = (0, slice(0, 2)),
+        >>>                 bottom_left=(1, 0))
+
+        >>> mg = MapsGrid(2, 2, ax_inits=ax_inits)
+        >>> mg.m_top.plot_map()
+        >>> mg.m_bottom_left.plot_map()
+
+        >>> mg.create_axes((1, 1), name="bottom_right")
+        >>> mg.ax_bottom_right.plot([1,2,3], [1,2,3])
+
+        """
+
+        if name is None:
+            # get all existing axes
+            axes = [key for key in self.__dict__ if key.startswith("ax_")]
+            name = str(len(axes))
+        else:
+            assert (
+                name.isidentifier()
+            ), f"the provided name {name} is not a valid identifier"
+
+        ax = self.f.add_subplot(self._gs[ax_init])
+        setattr(self, f"ax_{name}", ax)
+        return ax
 
     @property
     def children(self):
@@ -2102,11 +2354,14 @@ class MapsGrid:
         """
         self.parent.cb.click.share_events(*self.children)
 
-    def share_pick_events(self):
+    def share_pick_events(self, name="default"):
         """
         share pick events between all Maps objects of the grid
         """
-        self.parent.cb.pick.share_events(*self.children)
+        if name == "default":
+            self.parent.cb.pick.share_events(*self.children)
+        else:
+            self.parent.cb.pick[name].share_events(*self.children)
 
     def join_limits(self):
         """
